// mymalloc.h
// malloc wrapper to let me keep track of what is allocated

#ifndef __MYMALLOC_H
#define __MYMALLOC_H

#include "typ.h"        // bool
#include <stdio.h>      // NULL, FILE

class Malloc {
private:     // types
  // header per allocated block: doubly-linked list
  struct Header {
    Header *prev;      // previous, or NULL for first
    Header *next;      // next, or NULL for last
    int size;          // size of this block, in bytes, measured starting
                       // at &prev and ending at last user byte
    int userData;      // per-block data for use by traverseMemory
                       // (initially 0)

  public:
    // return the first byte of the user area
    void *userStart() const {
      return (void*)(this+1);
    }

    // return the last byte plus one of the user area
    void *userEnd() const {
      return (void*)((char*)this + size);
    }

    // return # of bytes user can safely use
    int userBytes() const {
      return size - sizeof(*this);
    }

    // return true if the given pointer is in the user space
    bool contains(void const *ptr) const {
      return userStart() <= ptr &&
                            ptr < userEnd();
    }

    // return if neighbors' links are consistent with my own
    bool locallyOk() const {
      return (!prev || prev->next == this) &&
             (!next || next->prev == this);
    }
  };

  class Index {
  public:
    int numBlocks;          // # of blocks allocated
    Header **array;         // sorted array of pointers to allocated blocks

  public:
    Index(int numBlocks);   // allocate array
    ~Index();               // deallocate array

    // binary search on array for block containing addr, NULL if none does
    Header *search(void const *addr) const;
  };

public:      // types
  struct Stats {
    int mallocs, frees, reallocs;    // calls to each
    int mallocdBytes, freedBytes;    // sums of bytes involved in the calls
    int freeAllNodes, freeAllBytes;  // stats for freeAll()

    int active() const { return mallocs - frees - freeAllNodes; }
    int activeBytes() const { return mallocdBytes - freedBytes - freeAllBytes; }
    void print(FILE *fp) const;      // write in human-consumable format
  };

private:     // data
  Header *first;     // first in the list; NULL when empty
  Stats stats;       // usage statistics
  Index *index;      // index of where things are allocated; NULL if not built
  bool changed;      // when true, index is invalid

private:     // funcs
  void buildIndex();

  // return header for given ptr-to-start-of-user-data
  Header const *getHeaderC(void const *ptr) const;
  Header *getHeader(void *ptr)
    { return (Header*)getHeaderC(ptr); }

public:      // funcs
  Malloc();
  ~Malloc();

  // ----------- ordinary user's interface -----------------
  // allocate memory with at least 'size' usable bytes
  // returns pointer to start of this memory
  void *malloc(unsigned size);

  // free memory allocated by malloc
  void free(void *ptr);

  // resize a block, preserving the contents possible to preserve
  void *realloc(void *ptr, unsigned newSize);


  // -------------- instrumentor's interface ---------------------
  // given a pointer:
  //   if it points somewhere into a user area,
  //     return the start of the user area
  //   else
  //     return NULL
  void *findBase(void const *possiblyInteriorPtr) const;

  // verify that the heap is ok, throwing an exception if it is not
  void checkHeap() const;

  // free all allocated blocks
  void freeAll();

  // return true if there are no allocated blocks
  bool isEmpty() const { return first == NULL; }

  // return the # of allocated blocks
  int numAllocated() const;

  // grab stats
  Stats const &getStats() const { return stats; }

  // get pointer to user-data for a specific block
  int *userData(void *ptr);

  // retrieve size of user's allocated data for a block
  unsigned allocatedSize(void const *ptr) const;

  // see if a block is a valid allocated block
  // NOTE: this fn sometimes segfaults on non-blocks, so it really should only
  //       be called when the expected answer is 'true' (I violate this a bit
  //       somewhere, but so far it hasn't caused a problem, so I'm not solving
  //       it (yet))
  bool isBlock(void const *ptr) const;

  // user-supplied fn to be called with each block
  //   start    - ptr to first byte
  //   end      - ptr to last byte plus one
  //   userData - ptr to per-block user-maintained info
  //   extra    - argument to traverseMemory
  //   returns  - true to continue, false to stop
  // note: this fn should *not* modify the Malloc object
  typedef bool (*TraversalFunc)(void *start, void *end, int *userData,
                                void *extra);

  // traverse all blocks until func returns false
  void traverseMemory(TraversalFunc func, void *extra);
};

#endif // __MYMALLOC_H

